#include "ShareableValue.h"
#include "SharedParent.h"
#include "NativeReanimatedModule.h"
#include "Logger.h"
#include "MutableValue.h"
#include "MutableValueSetterProxy.h"
#include "RemoteObject.h"
#include "FrozenObject.h"

namespace reanimated {

void ShareableValue::adaptCache(jsi::Runtime &rt, const jsi::Value &value) {
  // when adapting from host object we can assign cached value immediately such that we avoid
  // running `toJSValue` in the future when given object is accessed
  if (module->isUIRuntime(rt)) {
    remoteValue = std::make_unique<jsi::Value>(rt, value);
  } else {
    hostValue = std::make_unique<jsi::Value>(rt, value);
  }
}

void ShareableValue::adapt(jsi::Runtime &rt, const jsi::Value &value, ValueType objectType) {
  if (objectType == ValueType::MutableValueType) {
    type = ValueType::MutableValueType;
    mutableValue = std::make_shared<MutableValue>(rt, value, module);
  } else if (value.isUndefined()) {
    type = ValueType::UndefinedType;
  } else if (value.isNull()) {
    type = ValueType::NullType;
  } else if (value.isBool()) {
    type = ValueType::BoolType;
    boolValue = value.getBool();
  } else if (value.isNumber()) {
    type = ValueType::NumberType;
    numberValue = value.asNumber();
  } else if (value.isString()) {
    type = ValueType::StringType;
    stringValue = value.asString(rt).utf8(rt);
  } else if (value.isObject()) {
    auto object = value.asObject(rt);
    if (object.isFunction(rt)) {
      if (object.getProperty(rt, "__worklet").isUndefined()) {
        // not a worklet, we treat this as a host function
        type = ValueType::HostFunctionType;
        hostRuntime = &rt;
        hostFunction = std::make_shared<jsi::Function>(object.asFunction(rt));
      } else {
        // a worklet
        type = ValueType::WorkletFunctionType;
        frozenObject = std::make_shared<FrozenObject>(rt, object, module);
      }
    } else if (object.isArray(rt)) {
      type = ValueType::ArrayType;
      auto array = object.asArray(rt);
      for (size_t i = 0, size = array.size(rt); i < size; i++) {
        frozenArray.push_back(adapt(rt, array.getValueAtIndex(rt, i), module));
      }
    } else if (object.isHostObject<MutableValue>(rt)) {
      type = ValueType::MutableValueType;
      mutableValue = object.getHostObject<MutableValue>(rt);
      adaptCache(rt, value);
    } else if (object.isHostObject<RemoteObject>(rt)) {
      type = ValueType::RemoteObjectType;
      remoteObject = object.getHostObject<RemoteObject>(rt);
      adaptCache(rt, value);
    } else if (object.isHostObject<FrozenObject>(rt)) {
      type = ValueType::ObjectType;
      frozenObject = object.getHostObject<FrozenObject>(rt);
      adaptCache(rt, value);
    } else if (objectType == ValueType::RemoteObjectType) {
      type = ValueType::RemoteObjectType;
      remoteObject = std::make_shared<RemoteObject>(rt, object, module);
    } else {
      // create frozen object based on a copy of a given object
      type = ValueType::ObjectType;
      frozenObject = std::make_shared<FrozenObject>(rt, object, module);
    }
  } else if (value.isSymbol()) {
    type = ValueType::StringType;
    stringValue = value.asSymbol(rt).toString(rt);
  } else {
    throw "Invalid value type";
  }
}

std::shared_ptr<ShareableValue> ShareableValue::adapt(jsi::Runtime &rt, const jsi::Value &value, NativeReanimatedModule *module, ValueType valueType) {
  auto sv = std::shared_ptr<ShareableValue>(new ShareableValue(module));
  sv->adapt(rt, value, valueType);
  return sv;
}

jsi::Value ShareableValue::getValue(jsi::Runtime &rt) {
  // TODO: maybe we can cache toJSValue results on a per-runtime basis, need to avoid ref loops
  if (module->isUIRuntime(rt)) {
    if (remoteValue.get() == nullptr) {
      remoteValue = std::make_unique<jsi::Value>(rt, toJSValue(rt));
    }
    return jsi::Value(rt, *remoteValue);
  } else {
    if (hostValue.get() == nullptr) {
      hostValue = std::make_unique<jsi::Value>(rt, toJSValue(rt));
    }
    return jsi::Value(rt, *hostValue);
  }
  return toJSValue(rt);
}

jsi::Object ShareableValue::createHost(jsi::Runtime &rt, std::shared_ptr<jsi::HostObject> host) {
  return jsi::Object::createFromHostObject(rt, host);
}

jsi::Value ShareableValue::toJSValue(jsi::Runtime &rt) {
  switch (type) {
    case ValueType::UndefinedType:
      return jsi::Value::undefined();
    case ValueType::NullType:
      return jsi::Value::null();
    case ValueType::BoolType:
      return jsi::Value(boolValue);
    case ValueType::NumberType:
      return jsi::Value(numberValue);
    case ValueType::StringType:
      return jsi::Value(rt, jsi::String::createFromAscii(rt, stringValue));
    case ValueType::ObjectType:
      return createHost(rt, frozenObject);
    case ValueType::ArrayType: {
      jsi::Array array(rt, frozenArray.size());
      for (size_t i = 0; i < frozenArray.size(); i++) {
        array.setValueAtIndex(rt, i, frozenArray[i]->toJSValue(rt));
      }
      return array;
    }
    case ValueType::RemoteObjectType:
      if (module->isUIRuntime(rt)) {
        remoteObject->maybeInitializeOnUIRuntime(rt);
      }
      return createHost(rt, remoteObject);
    case ValueType::MutableValueType:
      return createHost(rt, mutableValue);
    case ValueType::HostFunctionType:
      if (hostRuntime == &rt) {
        // function is accessed from the same runtime it was crated, we just return same function obj
        return jsi::Value(rt, *hostFunction);
      } else {
        // function is accessed from a different runtme, we wrap function in host func that'd enqueue
        // call on an appropriate thread
        auto module = this->module;
        auto hostFunction = this->hostFunction;
        auto hostRuntime = this->hostRuntime;
        auto clb = [module, hostFunction, hostRuntime](
            jsi::Runtime &rt,
            const jsi::Value &thisValue,
            const jsi::Value *args,
            size_t count
            ) -> jsi::Value {
          // TODO: we should find thread based on runtime such that we could also call UI methods
          // from RN and not only RN methods from UI
          
          std::vector<std::shared_ptr<ShareableValue>> params;
          for (int i = 0; i < count; ++i) {
            params.push_back(ShareableValue::adapt(rt, args[i], module));
          }
          
          std::function<void()> job = [hostFunction, hostRuntime, params] {
            jsi::Value * args = new jsi::Value[params.size()];
            for (int i = 0; i < params.size(); ++i) {
              args[i] = params[i]->getValue(*hostRuntime);
            }
             
            jsi::Value returnedValue;
             
            returnedValue = hostFunction->call(*hostRuntime,
                                              static_cast<const jsi::Value*>(args),
                                              (size_t)params.size());
             
            delete [] args;
             // ToDo use returned value to return promise
          };
          
          module->scheduler->scheduleOnJS(job);
          return jsi::Value::undefined();
        };
        return jsi::Function::createFromHostFunction(rt, jsi::PropNameID::forAscii(rt, "hostFunction"), 0, clb);
      }
    case ValueType::WorkletFunctionType:
      auto module = this->module;
      auto frozenObject = this->frozenObject;
      if (module->isUIRuntime(rt)) {
        // when running on UI thread we prep a function

        auto jsThis = frozenObject->shallowClone(*module->runtime);
        std::shared_ptr<jsi::Function> funPtr(module->workletsCache->getFunction(rt, frozenObject));
        
        // HACK ALERT: there is a special case of "setter" function where we don't want to pass
        // closure as "this". Here we handle that case separately;
        if (module->valueSetter.get() == this) {
          return jsi::Value(rt, *funPtr);
        }

        auto clb = [=](
                   jsi::Runtime &rt,
                   const jsi::Value &thisValue,
                   const jsi::Value *args,
                   size_t count
                   ) mutable -> jsi::Value {
           jsi::Value oldJSThis = rt.global().getProperty(rt, "jsThis");
           rt.global().setProperty(rt, "jsThis", *jsThis); //set jsThis

           jsi::Value res = jsi::Value::undefined();
           try {
             if (thisValue.isObject()) {
               res = funPtr->callWithThis(rt, thisValue.asObject(rt), args, count);
             } else {
               res = funPtr->call(rt, args, count);
             }
           } catch(std::exception &e) {
               std::string str = e.what();
             this->module->errorHandler->setError(str);
             this->module->errorHandler->raise();
           }

           rt.global().setProperty(rt, "jsThis", oldJSThis); //clean jsThis
           return res;
        };
        return jsi::Function::createFromHostFunction(rt, jsi::PropNameID::forAscii(rt, "workletFunction"), 0, clb);
      } else {
        // when run outside of UI thread we enqueue a call on the UI thread
        auto retain_this = shared_from_this();
        auto clb = [retain_this = std::move(retain_this)](
            jsi::Runtime &rt,
            const jsi::Value &thisValue,
            const jsi::Value *args,
            size_t count
            ) -> jsi::Value {
          // TODO: we should find thread based on runtime such that we could also call UI methods
          // from RN and not only RN methods from UI
          
          auto module = retain_this->module;

          std::vector<std::shared_ptr<ShareableValue>> params;
          for (int i = 0; i < count; ++i) {
            params.push_back(ShareableValue::adapt(rt, args[i], module));
          }
          
          module->scheduler->scheduleOnUI([retain_this, params, &module] {
            jsi::Runtime &rt = *retain_this->module->runtime.get();
            auto jsThis = retain_this->createHost(rt, retain_this->frozenObject);
            auto code = jsThis.getProperty(rt, "asString").asString(rt).utf8(rt);
            std::shared_ptr<jsi::Function> funPtr(retain_this->module->workletsCache->getFunction(rt, retain_this->frozenObject));
            
            jsi::Value * args = new jsi::Value[params.size()];
            for (int i = 0; i < params.size(); ++i) {
              args[i] = params[i]->getValue(rt);
            }
            
            jsi::Value returnedValue;
            
            jsi::Value oldJSThis = rt.global().getProperty(rt, "jsThis");
            rt.global().setProperty(rt, "jsThis", jsThis); //set jsThis
            try {
              returnedValue = funPtr->call(rt,
                                             static_cast<const jsi::Value*>(args),
                                             (size_t)params.size());
            
            } catch(std::exception &e) {
              std::string str = e.what();
              module->errorHandler->setError(str);
              module->errorHandler->raise();
            }
            rt.global().setProperty(rt, "jsThis", oldJSThis); //clean jsThis
            
            delete [] args;
            // ToDo use returned value to return promise
          });
          return jsi::Value::undefined();
        };
        return jsi::Function::createFromHostFunction(rt, jsi::PropNameID::forAscii(rt, "workletFunction"), 0, clb);
      }
  }
  throw "convert error";
}

}
